import csv
import importlib
import json
import os
import sys
import types
import openpyxl
from typing import Tuple, Dict, Union, Text, List, Callable

import yaml
from loguru import logger
from pydantic import ValidationError

from rrunner import builtin, utils
from rrunner import exceptions
from rrunner.models import TestCase, ProjectMeta, TestSuite

from rrunner.common.handle_path import DIR



try:
    # PyYAML version >= 5.1
    # ref: https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
    yaml.warnings({"YAMLLoadWarning": False})
except AttributeError:
    pass


project_meta: Union[ProjectMeta, None] = None


def _load_excle_file(excle_file: Text) -> Dict:
    """ load excle file and check file content format
    """
    excleInfo = {}
    #excleInfoList = []
    # 获取工作簿对象
    wb = openpyxl.load_workbook(excle_file)

    #sheets = wb.get_sheet_names()
    sheets = wb.sheetnames

    excleInfo['name'] = os.path.basename(excle_file)
    excleInfo['sheetname'] = sheets[0]
    #excleInfoList.append(excleInfo)
    #testSteps = {}
    testcases = {}

    for i in range(len(sheets)):
        # 选择表单
        sh = wb[sheets[i]]
        # 按行获取表单所有格子中的数据，每一行的数据放在一个元组中
        res = list(sh.rows)
        # 获取第一行的数据，作为字典的键
        title = [item.value for item in res[0]]
        # 创建一个空列表，用来存放用例数据
        data_cases = []
        # 遍历除第一行之外的数据
        for item in res[1:]:
            # 获取该行数据的值
            data = [c.value for c in item]
            # 将该行数据和title（第一行数据）打包转换为字典
            data_case = dict(zip(title, data))
            # 将转换的字典添加到前面创建的空列表data_cases中

            if data_case['tag'] == 1:
                continue

            if data_case['validate'] == None:
                data_case['validate'] = []
            else:
                temp = data_case.pop('validate')
                if ';' in temp:
                    l = []
                    temp = temp.split(';')
                    for i in temp:
                        l.append(json.loads(i))
                    data_case['validate'] = l
                else:
                    data_case['validate'] = [json.loads(temp)]

            if data_case['setup_hooks'] == None:
                data_case['setup_hooks'] = []
            elif data_case['setup_hooks'].startswith("{"):
                data_case['setup_hooks'] = [json.loads(data_case.pop('setup_hooks'))]
            else:
                data_case['setup_hooks'] = [data_case.pop('setup_hooks')]

            if data_case['teardown_hooks'] == None:
                data_case['teardown_hooks'] = []
            elif data_case['teardown_hooks'].startswith("{"):
                data_case['teardown_hooks'] = [json.loads(data_case.pop('teardown_hooks'))]
            else:
                data_case['teardown_hooks'] = [data_case.pop('teardown_hooks')]

            # temp = json.loads(data_case.pop('variable'))
            # data_case['variables'] = temp

            if data_case['variables'] == None:
                data_case['variables'] = {}
            else:
                data_case['variables'] = json.loads(data_case.pop('variables'))

            if data_case['extract'] == None:
                data_case['extract'] = {}
            else:
                data_case['extract'] = json.loads(data_case.pop('extract'))

            newDic = {}
            newDic['method'] = data_case.pop('method').upper()
            newDic['contenttype'] = data_case.pop('contenttype')
            newDic['url'] = data_case.pop('url')

            #upload
            if data_case['upload'] != None:
                temp_upload = data_case.pop('upload')

                dict_upload = {}
                dict_upload['file'] = os.path.join(os.getcwd(), 'file', temp_upload)

                newDic['upload'] = dict_upload

            if newDic['method'] == 'GET':
                if data_case['data'] == None:
                    newDic['params'] = {}
                else:
                    newDic['params'] = json.loads(data_case.pop('data'))
            elif newDic['method'] == 'POST' and newDic['contenttype'] == 'json':
                if data_case['data'] == None:
                    newDic['json'] = {}
                else:
                    newDic['json'] = json.loads(data_case.pop('data'))
            elif newDic['method'] == 'POST' and newDic['contenttype'] == 'data':
                if data_case['data'] == None:
                    newDic['data'] = {}
                else:
                    newDic['data'] = json.loads(data_case.pop('data'))

            data_case['request'] = newDic

            data_cases.append(data_case)

    config = {'name': 'test'}
    testcases['teststeps'] = data_cases
    

    testcases['config'] = excleInfo

    return testcases


def load_test_file(test_file: Text) -> Dict:
    """load testcase/testsuite file content"""
    if not os.path.isfile(test_file):
        raise exceptions.FileNotFound(f"test file not exists: {test_file}")

    file_suffix = os.path.splitext(test_file)[1].lower()

    #load excle test file
    if file_suffix == ".xlsx":
        test_file_content = _load_excle_file(test_file)
    else:
        # '' or other suffix
        raise exceptions.FileFormatError(
            f"testcase/testsuite file should be YAML/JSON format, invalid format file: {test_file}"
        )

    return test_file_content


def load_testcase(testcase: Dict) -> TestCase:
    try:
        # validate with pydantic TestCase rrunel
        testcase_obj = TestCase.parse_obj(testcase)
    except ValidationError as ex:
        err_msg = f"TestCase ValidationError:\nerror: {ex}\ncontent: {testcase}"
        raise exceptions.TestCaseFormatError(err_msg)

    return testcase_obj


def load_testcase_file(testcase_file: Text) -> TestCase:
    """load testcase file and validate with pydantic rrunel"""
    testcase_content = load_test_file(testcase_file)
    testcase_obj = load_testcase(testcase_content)
    testcase_obj.config.path = testcase_file
    return testcase_obj


def load_folder_files(folder_path: Text, recursive: bool = True) -> List:
    """ load folder path, return all files endswith .yml/.yaml/.json/_test.py in list.

    Args:
        folder_path (str): specified folder path to load
        recursive (bool): load files recursively if True

    Returns:
        list: files endswith yml/yaml/json
    """
    if isinstance(folder_path, (list, set)):
        files = []
        for path in set(folder_path):
            files.extend(load_folder_files(path, recursive))

        return files

    if not os.path.exists(folder_path):
        return []

    file_list = []

    for dirpath, dirnames, filenames in os.walk(folder_path):
        filenames_list = []

        for filename in filenames:
            if not filename.lower().endswith((".yml", ".yaml", ".json",".xlsx", "_test.py")):
                continue

            filenames_list.append(filename)

        for filename in filenames_list:
            file_path = os.path.join(dirpath, filename)
            file_list.append(file_path)

        if not recursive:
            break

    return file_list


def load_module_functions(module) -> Dict[Text, Callable]:
    """ load python module functions.

    Args:
        module: python module

    Returns:
        dict: functions mapping for specified python module

            {
                "func1_name": func1,
                "func2_name": func2
            }

    """
    module_functions = {}

    for name, item in vars(module).items():
        if isinstance(item, types.FunctionType):
            module_functions[name] = item

    return module_functions


def load_builtin_functions() -> Dict[Text, Callable]:
    """ load builtin module functions
    """
    return load_module_functions(builtin)


def locate_file(start_path: Text, file_name: Text) -> Text:
    """ locate filename and return absolute file path.
        searching will be recursive upward until system root dir.

    Args:
        file_name (str): target locate file name
        start_path (str): start locating path, maybe file path or directory path

    Returns:
        str: located file path. None if file not found.

    Raises:
        exceptions.FileNotFound: If failed to locate file.

    """
    if os.path.isfile(start_path):
        start_dir_path = os.path.dirname(start_path)
    elif os.path.isdir(start_path):
        start_dir_path = start_path
    else:
        raise exceptions.FileNotFound(f"invalid path: {start_path}")

    file_path = os.path.join(start_dir_path, file_name)
    if os.path.isfile(file_path):
        # ensure absolute
        return os.path.abspath(file_path)

    # system root dir
    # Windows, e.g. 'E:\\'
    # Linux/Darwin, '/'
    parent_dir = os.path.dirname(start_dir_path)
    if parent_dir == start_dir_path:
        raise exceptions.FileNotFound(f"{file_name} not found in {start_path}")

    # locate recursive upward
    return locate_file(parent_dir, file_name)


def locate_debugtalk_py(start_path: Text) -> Text:
    """ locate debugtalk.py file

    Args:
        start_path (str): start locating path,
            maybe testcase file path or directory path

    Returns:
        str: debugtalk.py file path, None if not found

    """
    try:
        # locate debugtalk.py file.
        debugtalk_path = locate_file(start_path, "debugtalk.py")
    except exceptions.FileNotFound:
        debugtalk_path = None

    return debugtalk_path


def locate_project_root_directory(test_path: Text) -> Tuple[Text, Text]:
    """ locate debugtalk.py path as project root directory

    Args:
        test_path: specified testfile path

    Returns:
        (str, str): debugtalk.py path, project_root_directory

    """

    def prepare_path(path):
        if not os.path.exists(path):
            err_msg = f"path not exist: {path}"
            logger.error(err_msg)
            raise exceptions.FileNotFound(err_msg)

        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)

        return path

    test_path = prepare_path(test_path)

    # locate debugtalk.py file
    debugtalk_path = locate_debugtalk_py(test_path)

    if debugtalk_path:
        # The folder contains debugtalk.py will be treated as project RootDir.
        project_root_directory = os.path.dirname(debugtalk_path)
    else:
        # debugtalk.py not found, use os.getcwd() as project RootDir.
        project_root_directory = os.getcwd()

    return debugtalk_path, project_root_directory


def load_debugtalk_functions() -> Dict[Text, Callable]:
    """ load project debugtalk.py module functions
        debugtalk.py should be located in project root directory.

    Returns:
        dict: debugtalk module functions mapping
            {
                "func1_name": func1,
                "func2_name": func2
            }

    """
    # load debugtalk.py module
    try:
        imported_module = importlib.import_module("debugtalk")
    except Exception as ex:
        logger.error(f"error occurred in debugtalk.py: {ex}")
        sys.exit(1)

    # reload to refresh previously loaded module
    imported_module = importlib.reload(imported_module)
    return load_module_functions(imported_module)


def load_project_meta(test_path: Text, reload: bool = False) -> ProjectMeta:
    """ load testcases, .env, debugtalk.py functions.
        testcases folder is relative to project_root_directory
        by default, project_meta will be loaded only once, unless set reload to true.

    Args:
        test_path (str): test file/folder path, locate project RootDir from this path.
        reload: reload project meta if set true, default to false

    Returns:
        project loaded api/testcases definitions,
            environments and debugtalk.py functions.

    """
    global project_meta
    if project_meta and (not reload):
        return project_meta

    project_meta = ProjectMeta()

    if not test_path:
        return project_meta

    debugtalk_path, project_root_directory = locate_project_root_directory(test_path)

    # add project RootDir to sys.path
    sys.path.insert(0, project_root_directory)

    # load .env file
    # NOTICE:
    # environment variable maybe loaded in debugtalk.py
    # thus .env file should be loaded before loading debugtalk.py
    dot_env_path = os.path.join(project_root_directory, ".env")
    #dot_env = load_dot_env_file(dot_env_path)
    # if dot_env:
    #     project_meta.env = dot_env
    #     project_meta.dot_env_path = dot_env_path

    if debugtalk_path:
        # load debugtalk.py functions
        debugtalk_functions = load_debugtalk_functions()
    else:
        debugtalk_functions = {}

    # locate project RootDir and load debugtalk.py functions
    project_meta.RootDir = project_root_directory
    project_meta.functions = debugtalk_functions
    project_meta.debugtalk_path = debugtalk_path

    return project_meta


def convert_relative_project_root_dir(abs_path: Text) -> Text:
    """ convert absolute path to relative path, based on project_meta.RootDir

    Args:
        abs_path: absolute path

    Returns: relative path based on project_meta.RootDir

    """
    _project_meta = load_project_meta(abs_path)
    if not abs_path.startswith(_project_meta.RootDir):
        raise exceptions.ParamsError(
            f"failed to convert absolute path to relative path based on project_meta.RootDir\n"
            f"abs_path: {abs_path}\n"
            f"project_meta.RootDir: {_project_meta.RootDir}"
        )

    return abs_path[len(_project_meta.RootDir) + 1 :]
